"use strict";
/*!
 * Copyright (c) 2020 Ron Buckton (rbuckton@chronicles.org)
 *
 * This file is licensed to you under the terms of the MIT License, found in the LICENSE file
 * in the root of this repository or package.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.compareBaseline = exports.compareBaselines = exports.writeBaseline = exports.writeOutput = exports.writeNodes = exports.writeDiagnostics = exports.writeTokens = void 0;
const fs_1 = require("fs");
const path_1 = require("path");
const tokens_1 = require("../tokens");
function writeTokens(test, scanner, lineMap, baselines) {
    let text = `/// ${test}:\r\n`;
    let token;
    do {
        token = scanner.scan();
        let message = `SyntaxKind[${(0, tokens_1.formatKind)(token)}](${lineMap.formatOffset(scanner.getTokenPos())}): `;
        switch (token) {
            case tokens_1.SyntaxKind.ProseFull:
            case tokens_1.SyntaxKind.ProseHead:
            case tokens_1.SyntaxKind.ProseMiddle:
            case tokens_1.SyntaxKind.ProseTail:
                message += scanner.getTokenValue();
                break;
            case tokens_1.SyntaxKind.Identifier:
                message += `${scanner.getTokenValue()}`;
                break;
            case tokens_1.SyntaxKind.TerminalLiteral:
                message += `\`${scanner.getTokenValue()}\``;
                break;
            case tokens_1.SyntaxKind.StringLiteral:
                message += `"${scanner.getTokenValue()}"`;
                break;
            case tokens_1.SyntaxKind.UnicodeCharacterLiteral:
                message += scanner.getTokenText();
                break;
            default:
                message += `${(0, tokens_1.tokenToString)(token)}`;
        }
        text += message + "\r\n";
    } while (token !== tokens_1.SyntaxKind.EndOfFileToken);
    return writeBaseline(test + ".tokens", text, baselines);
}
exports.writeTokens = writeTokens;
function writeDiagnostics(test, diagnostics, baselines) {
    let text = undefined;
    diagnostics.forEach(message => {
        if (!text) {
            text = `/// ${test}:\r\n`;
        }
        text += message + "\r\n";
    });
    return writeBaseline(test + ".diagnostics", text, baselines);
}
exports.writeDiagnostics = writeDiagnostics;
function writeNodes(test, sourceFile, baselines) {
    let text = `/// ${test}:\r\n`;
    let indents = ["", "  "];
    let indentDepth = 0;
    printNode(sourceFile);
    return writeBaseline(test + ".nodes", text, baselines);
    function getIndent(depth) {
        if (depth >= indents.length) {
            indents[depth] = getIndent(depth - 1) + indents[1];
        }
        return indents[depth];
    }
    function printNode(node) {
        text += getIndent(indentDepth) + formatNode(node, sourceFile) + "\r\n";
        indentDepth++;
        for (const child of node.children()) {
            printNode(child);
        }
        indentDepth--;
    }
}
exports.writeNodes = writeNodes;
function writeOutput(test, extname, text, baselines) {
    return writeBaseline(test + extname, text, baselines);
}
exports.writeOutput = writeOutput;
function writeBaseline(file, text, baselines) {
    if (baselines) {
        baselines.push(file);
    }
    let { localFile } = resolveBaseline(file);
    if (text === undefined) {
        if ((0, fs_1.existsSync)(localFile)) {
            (0, fs_1.unlinkSync)(localFile);
        }
    }
    else {
        (0, fs_1.writeFileSync)(localFile, text, { encoding: "utf8" });
    }
    return file;
}
exports.writeBaseline = writeBaseline;
function compareBaselines(baselines) {
    for (let file of baselines) {
        let { localFile, referenceFile } = resolveBaseline(file);
        let localText = (0, fs_1.existsSync)(localFile) ? (0, fs_1.readFileSync)(localFile, "utf8") : undefined;
        let referenceText = (0, fs_1.existsSync)(referenceFile) ? (0, fs_1.readFileSync)(referenceFile, "utf8") : undefined;
        if (localText !== referenceText) {
            throw new Error(`The baseline file '${file}' has changed.`);
        }
    }
}
exports.compareBaselines = compareBaselines;
function compareBaseline(file) {
    let { localFile, referenceFile } = resolveBaseline(file);
    let localText = (0, fs_1.existsSync)(localFile) ? (0, fs_1.readFileSync)(localFile, "utf8") : undefined;
    let referenceText = (0, fs_1.existsSync)(referenceFile) ? (0, fs_1.readFileSync)(referenceFile, "utf8") : undefined;
    if (localText !== referenceText) {
        throw new Error(`The baseline file '${file}' has changed.`);
    }
}
exports.compareBaseline = compareBaseline;
function resolveBaseline(file) {
    let baselinePath = (0, path_1.resolve)(__dirname, "../../baselines");
    let localPath = (0, path_1.resolve)(baselinePath, "local");
    let localFile = (0, path_1.resolve)(localPath, file);
    let referencePath = (0, path_1.resolve)(baselinePath, "reference");
    let referenceFile = (0, path_1.resolve)(referencePath, file);
    try {
        (0, fs_1.mkdirSync)((0, path_1.dirname)(localFile), { recursive: true });
    }
    catch (_a) { }
    try {
        (0, fs_1.mkdirSync)((0, path_1.dirname)(referenceFile), { recursive: true });
    }
    catch (_b) { }
    return { localFile, referenceFile };
}
function formatNode(node, sourceFile) {
    var text = `(${sourceFile.lineMap.formatOffset(node.getStart(sourceFile))})`;
    text += `SyntaxKind[${(0, tokens_1.formatKind)(node.kind)}]`;
    switch (node.kind) {
        case tokens_1.SyntaxKind.Identifier:
        case tokens_1.SyntaxKind.TerminalLiteral:
        case tokens_1.SyntaxKind.ProseFull:
        case tokens_1.SyntaxKind.ProseHead:
        case tokens_1.SyntaxKind.ProseMiddle:
        case tokens_1.SyntaxKind.ProseTail:
        case tokens_1.SyntaxKind.StringLiteral:
            text += `(text = "${node.text}")`;
            break;
        case tokens_1.SyntaxKind.UnicodeCharacterLiteral:
            text += `(text = ${sourceFile.text.slice(node.getStart(sourceFile), node.end)})`;
            break;
        case tokens_1.SyntaxKind.SourceFile:
            text += `(filename = "${(0, path_1.basename)(node.filename)}")`;
            break;
    }
    switch (node.kind) {
        case tokens_1.SyntaxKind.Terminal:
        case tokens_1.SyntaxKind.Nonterminal:
            if (node.questionToken) {
                text += "?";
            }
            break;
        case tokens_1.SyntaxKind.Argument:
            if (node.operatorToken) {
                switch (node.operatorToken.kind) {
                    case tokens_1.SyntaxKind.QuestionToken:
                        text += "?";
                        break;
                    case tokens_1.SyntaxKind.TildeToken:
                        text += "~";
                        break;
                    case tokens_1.SyntaxKind.PlusToken:
                        text += "+";
                        break;
                }
            }
            break;
    }
    return text;
}
//# sourceMappingURL=diff.js.map